@@ -22,7 +22,7 @@
 # along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.
 
 """
-Copyright 2014-2020 Yevhenii Prokopalo, Anthony Larcher
+Copyright 2014-2021 Anthony Larcher, Yevhenii Prokopalo
 """
 
 
@@ -31,18 +31,21 @@ import traceback
 import logging
 import matplotlib.pyplot as plt
 import multiprocessing
+import os
 import numpy
 import pandas
 import pickle
 import shutil
+import sys
 import time
 import torch
-import torch.optim as optim
+import tqdm
 import yaml
 
 from torchvision import transforms
 from collections import OrderedDict
 from .xsets import SideSet
+from .xsets import FileSet
 from .xsets import IdMapSet
 from .res_net import RawPreprocessor, ResBlockWFMS
 from ..bosaris import IdMap
@@ -50,14 +53,17 @@ from ..statserver import StatServer
 from torch.utils.data import DataLoader
 from sklearn.model_selection import train_test_split
 from .sincnet import SincNet
-#from torch.utils.tensorboard import SummaryWriter
 from .loss import ArcLinear
+from .loss import ArcFace
+from .loss import l2_norm
 
-import tqdm
+
+
+os.environ['MKL_THREADING_LAYER'] = 'GNU'
 
 __license__ = "LGPL"
 __author__ = "Anthony Larcher"
-__copyright__ = "Copyright 2015-2020 Anthony Larcher"
+__copyright__ = "Copyright 2015-2021 Anthony Larcher"
 __maintainer__ = "Anthony Larcher"
 __email__ = "anthony.larcher@univ-lemans.fr"
 __status__ = "Production"
@@ -94,13 +100,6 @@ class GuruMeditation (torch.autograd.detect_anomaly):
         pdb.set_trace()
 
 
-
-
-
-
-
-
-
 def select_n_random(data, labels, n=100):
     '''
     Selects n random datapoints and their corresponding labels from a dataset
@@ -438,6 +437,11 @@ class Xtractor(torch.nn.Module):
                                                                 dilation=cfg["segmental"][k]["dilation"])))
                     input_size = cfg["segmental"][k]["output_channels"]
 
+                elif k.startswith("ctrans"):
+                    segmental_layers.append((k, torch.nn.ConvTranspose1d(input_size,
+                                                                         cfg["segmental"][k]["output_channels"],
+                                                                         kernel_size=cfg["segmental"][k]["kernel_size"],
+                                                                         dilation=cfg["segmental"][k]["dilation"])))
                 elif k.startswith("activation"):
                     segmental_layers.append((k, self.activation))
 
@@ -519,10 +523,14 @@ class Xtractor(torch.nn.Module):
 
             elif self.loss == "aam":
                 self.norm_embedding = True
-                self.after_speaker_embedding = ArcLinear(input_size,
-                                                         self.speaker_number,
-                                                         margin=self.aam_margin,
-                                                         s=self.aam_s)
+                #self.after_speaker_embedding = ArcLinear(input_size,
+                #                                         self.speaker_number,
+                #                                         margin=self.aam_margin,
+                #                                         s=self.aam_s)
+                self.after_speaker_embedding = ArcFace(embedding_size=input_size,
+                                                       classnum=self.speaker_number,
+                                                       s=64.,
+                                                       m=0.5)
 
             self.after_speaker_embedding_weight_decay = cfg["after_embedding"]["weight_decay"]
 
@@ -545,8 +553,10 @@ class Xtractor(torch.nn.Module):
         x = self.before_speaker_embedding(x)
 
         if self.norm_embedding:
-            x_norm = x.norm(p=2,dim=1, keepdim=True) / 10. # Why  10. ?
-            x = torch.div(x, x_norm)
+            #x_norm = x.norm(p=2,dim=1, keepdim=True) / 10. # Why  10. ?
+            #x_norm = torch.linalg.norm(x, ord=2, dim=1, keepdim=True, out=None, dtype=None)
+            #x = torch.div(x, x_norm)
+            x = l2_norm(x)
 
         if is_eval:
             return x
@@ -591,7 +601,10 @@ def xtrain(speaker_number,
            opt=None,
            reset_parts=[],
            freeze_parts=[],
-           num_thread=None):
+           num_thread=None,
+           write_batches_to_disk=False,
+           load_batches_from_disk=False,
+           tmp_batch_dir=None):
     """
 
     :param speaker_number:
@@ -619,13 +632,16 @@ def xtrain(speaker_number,
     #writer = SummaryWriter("runs/xvectors_experiments_2")
     writer = None
 
+    if write_batches_to_disk:
+        load_batches_from_disk = True
+
     if num_thread is None:
         num_thread = multiprocessing.cpu_count()
 
+    logging.critical(f"Use {num_thread} cpus")
     logging.critical(f"Start process at {time.strftime('%H:%M:%S', time.localtime())}")
 
     device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
-
     # Start from scratch
     if model_name is None and model_yaml in ["xvector", "rawnet2"]:
         # Initialize a first model
@@ -691,45 +707,75 @@ def xtrain(speaker_number,
     if torch.cuda.device_count() > 1 and multi_gpu:
         print("Let's use", torch.cuda.device_count(), "GPUs!")
         model = torch.nn.DataParallel(model)
+        #model = DDP(model)
     else:
         print("Train on a single GPU")
     model.to(device)
 
-    """
-    Set the dataloaders according to the dataset_yaml
-    
-    First we load the dataframe from CSV file in order to split it for training and validation purpose
-    Then we provide those two 
-    """
     with open(dataset_yaml, "r") as fh:
         dataset_params = yaml.load(fh, Loader=yaml.FullLoader)
         df = pandas.read_csv(dataset_params["dataset_description"])
-    training_df, validation_df = train_test_split(df, test_size=dataset_params["validation_ratio"])
-    torch.manual_seed(dataset_params['seed'])
-    training_set = SideSet(dataset_yaml, 
-                           set_type="train", 
-                           dataset_df=training_df, 
-                           chunk_per_segment=dataset_params['train']['chunk_per_segment'], 
-                           overlap=dataset_params['train']['overlap'])
+
+    if load_batches_from_disk:
+        train_batch_fn_format = tmp_batch_dir + "/train/train_{}_batch.h5"
+        val_batch_fn_format = tmp_batch_dir + "/val/val_{}_batch.h5"
+
+    if not load_batches_from_disk or write_batches_to_disk:
+        """
+        Set the dataloaders according to the dataset_yaml
+        
+        First we load the dataframe from CSV file in order to split it for training and validation purpose
+        Then we provide those two 
+        """
+
+        if write_batches_to_disk or dataset_params["batch_size"] > 1:
+            output_format = "numpy"
+        else:
+            output_format = "pytorch"
+
+        training_df, validation_df = train_test_split(df, test_size=dataset_params["validation_ratio"])
+        torch.manual_seed(dataset_params['seed'])
+        training_set = SideSet(dataset_yaml,
+                               set_type="train",
+                               dataset_df=training_df,
+                               chunk_per_segment=dataset_params['train']['chunk_per_segment'],
+                               overlap=dataset_params['train']['overlap'],
+                               output_format=output_format)
+
+        validation_set = SideSet(dataset_yaml,
+                                 set_type="validation",
+                                 dataset_df=validation_df,
+                                 output_format=output_format)
+
+
+        if write_batches_to_disk:
+            logging.critical("Start writing batches on disk")
+            training_set.write_to_disk(dataset_params["batch_size"], train_batch_fn_format, num_thread)
+            validation_set.write_to_disk(dataset_params["batch_size"], val_batch_fn_format, num_thread)
+            logging.critical("---> Done")
+
+    if load_batches_from_disk:
+        training_set = FileSet(train_batch_fn_format)
+        validation_set = FileSet(train_batch_fn_format)
+        batch_size = 1
+    else:
+        batch_size = dataset_params["batch_size"]
+
+
+    print(f"Size of batches = {batch_size}")
     training_loader = DataLoader(training_set,
-                                 batch_size=dataset_params["batch_size"],
+                                 batch_size=batch_size,
                                  shuffle=True,
                                  drop_last=True,
                                  pin_memory=True,
                                  num_workers=num_thread)
 
-    validation_set = SideSet(dataset_yaml, set_type="validation", dataset_df=validation_df)
     validation_loader = DataLoader(validation_set,
-                                   batch_size=dataset_params["batch_size"],
+                                   batch_size=batch_size,
                                    drop_last=True,
                                    pin_memory=True,
                                    num_workers=num_thread)
 
-    # Add for TensorBoard
-    #dataiter = iter(training_loader)
-    #data, labels = dataiter.next()
-    #writer.add_graph(model, data)
-
 
     """
     Set the training options
@@ -744,19 +790,19 @@ def xtrain(speaker_number,
         _optimizer = torch.optim.SGD
         _options = {'lr': lr, 'momentum': 0.9}
 
-    params = [
-        {
-            'params': [
-                param for name, param in model.named_parameters() if 'bn' not in name
-            ]
-        },
-        {
-            'params': [
-                param for name, param in model.named_parameters() if 'bn' in name
-            ],
-            'weight_decay': 0
-        },
-    ]
+    #params = [
+    #    {
+    #        'params': [
+    #            param for name, param in model.named_parameters() if 'bn' not in name
+    #        ]
+    #    },
+    #    {
+    #        'params': [
+    #            param for name, param in model.named_parameters() if 'bn' in name
+    #        ],
+    #        'weight_decay': 0
+    #    },
+    #]
 
     param_list = []
     if type(model) is Xtractor:
@@ -870,14 +916,15 @@ def train_epoch(model, epoch, training_loader, optimizer, log_interval, device,
     accuracy = 0.0
     running_loss = 0.0
     for batch_idx, (data, target) in enumerate(training_loader):
+        data = data.squeeze().to(device)
         target = target.squeeze()
         target = target.to(device)
         optimizer.zero_grad()
 
         if loss_criteria == 'aam':
-            output = model(data.to(device), target=target)
+            output = model(data, target=target)
         else:
-            output = model(data.to(device), target=None)
+            output = model(data, target=None)
 
         #with GuruMeditation():
         loss = criterion(output, target)
@@ -892,7 +939,8 @@ def train_epoch(model, epoch, training_loader, optimizer, log_interval, device,
             accuracy += (torch.argmax(output.data, 1) == target).sum()
             if batch_idx % log_interval == 0:
                 batch_size = target.shape[0]
-                logging.critical('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAccuracy: {:.3f}'.format(
+                logging.critical('{}, Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAccuracy: {:.3f}'.format(
+                    time.strftime('%H:%M:%S', time.localtime()),
                     epoch, batch_idx + 1, training_loader.__len__(),
                     100. * batch_idx / training_loader.__len__(), loss.item(),
                     100.0 * accuracy.item() / ((batch_idx + 1) * batch_size)))
@@ -946,16 +994,17 @@ def cross_validation(model, validation_loader, device):
     with torch.no_grad():
         for batch_idx, (data, target) in enumerate(validation_loader):
             batch_size = target.shape[0]
-            target = target.squeeze()
+            target = target.squeeze().to(device)
+            data = data.squeeze().to(device) 
 
             if loss_criteria == "aam":
-                output = model(data.to(device), target=target)
+                output = model(data, target=target)
             else:
-                output = model(data.to(device), target=None)
+                output = model(data, target=None)
 
-            accuracy += (torch.argmax(output.data, 1) == target.to(device)).sum()
+            accuracy += (torch.argmax(output.data, 1) == target).sum()
 
-            loss += criterion(output, target.to(device))
+            loss += criterion(output, target)
 
     return 100. * accuracy.cpu().numpy() / ((batch_idx + 1) * batch_size), \
            loss.cpu().numpy() / ((batch_idx + 1) * batch_size)
@@ -1028,6 +1077,8 @@ def extract_embeddings(idmap_name,
         # Process the data
         with torch.no_grad():
             for idx, (data, mod, seg, start, stop) in tqdm.tqdm(enumerate(dataloader)):
+                if data.shape[1] > 20000000:
+                    data = data[...,:20000000]
                 vec = model(data.to(device), is_eval=True)
                 embeddings.stat1[idx, :] = vec.detach().cpu()
 
@@ -1089,3 +1140,308 @@ def extract_sliding_embedding(idmap_name,
 
     return embeddings
 
+
+def xdebug(speaker_number,
+           dataset_yaml,
+           epochs=None,
+           lr=None,
+           model_yaml=None,
+           model_name=None,
+           loss=None,
+           aam_margin=None,
+           aam_s=None,
+           patience=None,
+           tmp_model_name=None,
+           best_model_name=None,
+           multi_gpu=True,
+           clipping=False,
+           opt=None,
+           reset_parts=[],
+           freeze_parts=[],
+           num_thread=None):
+    """
+
+    :param speaker_number:
+    :param dataset_yaml:
+    :param epochs:
+    :param lr:
+    :param model_yaml:
+    :param model_name:
+    :param loss:
+    :param aam_margin:
+    :param aam_s:
+    :param patience:
+    :param tmp_model_name:
+    :param best_model_name:
+    :param multi_gpu:
+    :param clipping:
+    :param opt:
+    :param reset_parts:
+    :param freeze_parts:
+    :param num_thread:
+    :return:
+    """
+    # Add for tensorboard
+    # Writer will output to ./runs/ directory by default
+    # writer = SummaryWriter("runs/xvectors_experiments_2")
+    writer = None
+
+    if num_thread is None:
+        num_thread = multiprocessing.cpu_count()
+
+    logging.critical(f"Start process at {time.strftime('%H:%M:%S', time.localtime())}")
+
+    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
+
+    # Start from scratch
+    if model_name is None and model_yaml in ["xvector", "rawnet2"]:
+        # Initialize a first model
+        if model_yaml == "xvector":
+            model = Xtractor(speaker_number, "xvector")
+        elif model_yaml == "rawnet2":
+            model = Xtractor(speaker_number, "rawnet2")
+    else:
+        with open(model_yaml, 'r') as fh:
+            model_archi = yaml.load(fh, Loader=yaml.FullLoader)
+            if epochs is None:
+                epochs = model_archi["training"]["epochs"]
+            if patience is None:
+                patience = model_archi["training"]["patience"]
+            if opt is None:
+                opt = model_archi["training"]["opt"]
+            if lr is None:
+                lr = model_archi["training"]["lr"]
+            if loss is None:
+                loss = model_archi["training"]["loss"]
+            if aam_margin is None and model_archi["training"]["loss"] == "aam":
+                aam_margin = model_archi["training"]["aam_margin"]
+            if aam_s is None and model_archi["training"]["loss"] == "aam":
+                aam_s = model_archi["training"]["aam_s"]
+            if tmp_model_name is None:
+                tmp_model_name = model_archi["training"]["tmp_model_name"]
+            if best_model_name is None:
+                best_model_name = model_archi["training"]["best_model_name"]
+            if multi_gpu is None:
+                multi_gpu = model_archi["training"]["multi_gpu"]
+            if clipping is None:
+                clipping = model_archi["training"]["clipping"]
+
+        if model_name is None:
+            model = Xtractor(speaker_number, model_yaml)
+
+        # If we start from an existing model
+        else:
+            # Load the model
+            logging.critical(f"*** Load model from = {model_name}")
+            checkpoint = torch.load(model_name)
+            model = Xtractor(speaker_number, model_yaml)
+
+            """
+            Here we remove all layers that we don't want to reload
+
+            """
+            pretrained_dict = checkpoint["model_state_dict"]
+            for part in reset_parts:
+                pretrained_dict = {k: v for k, v in pretrained_dict.items() if not k.startswith(part)}
+
+            new_model_dict = model.state_dict()
+            new_model_dict.update(pretrained_dict)
+            model.load_state_dict(new_model_dict)
+
+        # Freeze required layers
+        for name, param in model.named_parameters():
+            if name.split(".")[0] in freeze_parts:
+                param.requires_grad = False
+
+    print(model)
+
+    if torch.cuda.device_count() > 1 and multi_gpu:
+        print("Let's use", torch.cuda.device_count(), "GPUs!")
+        model = torch.nn.DataParallel(model)
+    else:
+        print("Train on a single GPU")
+    model.to(device)
+
+    """
+    Set the dataloaders according to the dataset_yaml
+
+    First we load the dataframe from CSV file in order to split it for training and validation purpose
+    Then we provide those two 
+    """
+    with open(dataset_yaml, "r") as fh:
+        dataset_params = yaml.load(fh, Loader=yaml.FullLoader)
+        df = pandas.read_csv(dataset_params["dataset_description"])
+    training_df, validation_df = train_test_split(df, test_size=dataset_params["validation_ratio"])
+    torch.manual_seed(dataset_params['seed'])
+    training_set = SideSet(dataset_yaml,
+                           set_type="train",
+                           dataset_df=training_df,
+                           chunk_per_segment=dataset_params['train']['chunk_per_segment'],
+                           overlap=dataset_params['train']['overlap'],
+                           output_format="numpy")
+
+    validation_set = SideSet(dataset_yaml, set_type="validation", dataset_df=validation_df)
+
+    return model, training_set, validation_set
+
+def xtime(model, training_set, validation_set,
+          speaker_number,
+          model_yaml,
+          batch_size,
+          num_thread,
+          ):
+    training_loader = DataLoader(training_set,
+                                 batch_size=batch_size,
+                                 shuffle=True,
+                                 drop_last=True,
+                                 pin_memory=True,
+                                 num_workers=num_thread)
+
+    validation_loader = DataLoader(validation_set,
+                                   batch_size=batch_size,
+                                   drop_last=True,
+                                   pin_memory=True,
+                                   num_workers=num_thread)
+    """
+    Set the training options
+    """
+    patience = 10
+    epochs = 2
+    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
+    with open(model_yaml, 'r') as fh:
+        model_archi = yaml.load(fh, Loader=yaml.FullLoader)
+        epochs = model_archi["training"]["epochs"]
+        patience = model_archi["training"]["patience"]
+        opt = model_archi["training"]["opt"]
+        lr = model_archi["training"]["lr"]
+        loss = model_archi["training"]["loss"]
+        aam_margin = model_archi["training"]["aam_margin"]
+        aam_s = model_archi["training"]["aam_s"]
+        tmp_model_name = model_archi["training"]["tmp_model_name"]
+        best_model_name = model_archi["training"]["best_model_name"]
+        multi_gpu = model_archi["training"]["multi_gpu"]
+        clipping = model_archi["training"]["clipping"]
+
+    if opt == 'adam':
+        _optimizer = torch.optim.Adam
+        _options = {'lr': lr}
+    elif opt == 'rmsprop':
+        _optimizer = torch.optim.RMSprop
+        _options = {'lr': lr}
+    else:  # opt == 'sgd'
+        _optimizer = torch.optim.SGD
+        _options = {'lr': lr, 'momentum': 0.9}
+
+    params = [
+        {
+            'params': [
+                param for name, param in model.named_parameters() if 'bn' not in name
+            ]
+        },
+        {
+            'params': [
+                param for name, param in model.named_parameters() if 'bn' in name
+            ],
+            'weight_decay': 0
+        },
+    ]
+
+    param_list = []
+    if type(model) is Xtractor:
+        if model.preprocessor is not None:
+            param_list.append(
+                {'params': model.preprocessor.parameters(), 'weight_decay': model.preprocessor_weight_decay})
+        param_list.append(
+            {'params': model.sequence_network.parameters(), 'weight_decay': model.sequence_network_weight_decay})
+        param_list.append({'params': model.stat_pooling.parameters(), 'weight_decay': model.stat_pooling_weight_decay})
+        param_list.append({'params': model.before_speaker_embedding.parameters(),
+                           'weight_decay': model.before_speaker_embedding_weight_decay})
+        param_list.append({'params': model.after_speaker_embedding.parameters(),
+                           'weight_decay': model.after_speaker_embedding_weight_decay})
+
+    else:
+        if model.module.preprocessor is not None:
+            param_list.append({'params': model.module.preprocessor.parameters(),
+                               'weight_decay': model.module.preprocessor_weight_decay})
+        param_list.append({'params': model.module.sequence_network.parameters(),
+                           'weight_decay': model.module.sequence_network_weight_decay})
+        param_list.append(
+            {'params': model.module.stat_pooling.parameters(), 'weight_decay': model.module.stat_pooling_weight_decay})
+        param_list.append({'params': model.module.before_speaker_embedding.parameters(),
+                           'weight_decay': model.module.before_speaker_embedding_weight_decay})
+        param_list.append({'params': model.module.after_speaker_embedding.parameters(),
+                           'weight_decay': model.module.after_speaker_embedding_weight_decay})
+
+    optimizer = _optimizer(param_list, **_options)
+
+    # optimizer = torch.optim.SGD(params,
+    #                            lr=lr,
+    #                            momentum=0.9,
+    #                            weight_decay=0.0005)
+    # print(f"Learning rate = {lr}")
+
+    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', verbose=True)
+
+    best_accuracy = 0.0
+    best_accuracy_epoch = 1
+    curr_patience = patience
+    for epoch in range(1, epochs + 1):
+        # Process one epoch and return the current model
+        if curr_patience == 0:
+            print(f"Stopping at epoch {epoch} for cause of patience")
+            break
+        model = train_epoch(model,
+                            epoch,
+                            training_loader,
+                            optimizer,
+                            10,
+                            device=device,
+                            clipping=clipping,
+                            tb_writer=None)
+
+        # Add the cross validation here
+        accuracy, val_loss = cross_validation(model, validation_loader, device=device)
+        logging.critical(f"***{time.strftime('%H:%M:%S', time.localtime())} Cross validation accuracy = {accuracy} %")
+
+        # Decrease learning rate according to the scheduler policy
+        scheduler.step(val_loss)
+
+        # remember best accuracy and save checkpoint
+        is_best = accuracy > best_accuracy
+        best_accuracy = max(accuracy, best_accuracy)
+
+        if type(model) is Xtractor:
+            save_checkpoint({
+                'epoch': epoch,
+                'model_state_dict': model.state_dict(),
+                'optimizer_state_dict': optimizer.state_dict(),
+                'accuracy': best_accuracy,
+                'scheduler': scheduler,
+                'speaker_number': speaker_number,
+                'model_archi': model_archi
+            }, is_best, filename=tmp_model_name + ".pt", best_filename=best_model_name + '.pt')
+        else:
+            save_checkpoint({
+                'epoch': epoch,
+                'model_state_dict': model.module.state_dict(),
+                'optimizer_state_dict': optimizer.state_dict(),
+                'accuracy': best_accuracy,
+                'scheduler': scheduler,
+                'speaker_number': speaker_number,
+                'model_archi': model_archi
+            }, is_best, filename=tmp_model_name + ".pt", best_filename=best_model_name + '.pt')
+
+        if is_best:
+            best_accuracy_epoch = epoch
+            curr_patience = patience
+        else:
+            curr_patience -= 1
+    # writer.close()
+
+    for ii in range(torch.cuda.device_count()):
+        print(torch.cuda.memory_summary(ii))
+
+    logging.critical(f"Best accuracy {best_accuracy * 100.} obtained at epoch {best_accuracy_epoch}")
+
+
+
